/*!
 * \file
 * \brief   Distributed sort implementation header
 *
 * \author
 *    Christos Choutouridis AEM:8997
 *    <cchoutou@ece.auth.gr>
 */

#ifndef DISTBITONIC_H_
#define DISTBITONIC_H_

#include <vector>
#include <algorithm>
#include <parallel/algorithm>
#include <cmath>
#include <cstdint>
#if !defined DEBUG
#define NDEBUG
#endif
#include <cassert>

#include "utils.hpp"

/*
 * Exported timers
 */
extern Timing Timer_total;
extern Timing Timer_fullSort;
extern Timing Timer_exchange;
extern Timing Timer_minmax;
extern Timing Timer_elbowSort;



/*
 * ============================== Sort utilities ==============================
 */



/*!
 * Returns the ascending or descending configuration of the node's sequence based on
 * the current node (MPI process) and the depth of the sorting network
 *
 * @param node      [mpi_id_t] The current node (MPI process)
 * @param depth     [size_t]   The total depth of the sorting network (same for each step for a given network)
 * @return          [bool]     True if we need ascending configuration, false otherwise
 */
bool ascending(mpi_id_t node, size_t depth);

/*!
 * Returns the node's partner for data exchange during the sorting network iterations
 * of Bitonic
 *
 * @param node      [mpi_id_t] The current node
 * @param step      [size_t]   The step of the sorting network
 * @return          [mpi_id_t] The node id of the partner for data exchange
 */
mpi_id_t partner(mpi_id_t node, size_t step);


/*!
 * Predicate to check if a node keeps the small numbers during the bitonic sort network exchange.
 *
 * @param node      [mpi_id_t] The node for which we check
 * @param partner   [mpi_id_t] The partner of the data exchange
 * @param depth     [size_t]   The total depth of the sorting network (same for each step for a given network)
 * @return          [bool]     True if the node should keep the small values, false otherwise
 */
bool keepSmall(mpi_id_t node, mpi_id_t partner, size_t depth);



/*
 * ============================== Data utilities ==============================
 */

/*!
 * Sort a range using the build-in O(Nlog(N)) algorithm
 *
 * @tparam RangeT   A range type with random access iterator
 *
 * @param data      [RangeT] The data to be sorted
 * @param ascending [bool]   Flag to indicate the sorting order
 */
template<typename RangeT>
void fullSort(RangeT& data, bool ascending) noexcept {
    // Use introsort from stdlib++ here, unless ... __gnu_parallel
    if (ascending) {
        __gnu_parallel::sort(data.begin(), data.end(), std::less<>());
    }
    else {
        __gnu_parallel::sort(data.begin(), data.end(), std::greater<>());
    }
}

/*!
 * Core functionality of sort for shadowed buffer types using
 * the "elbow sort" algorithm.
 *
 * @note:
 *  This algorithm can not work "in place".
 *  We use the active buffer as source and the shadow as target.
 *  At the end we switch which buffer is active and which is the shadow.
 * @note
 *  This is the core functionality. Use the elbowSort() function instead
 *
 * @tparam ShadowedDataT    A Shadowed buffer type with random access iterator.
 * @tparam CompT            A Comparison type for binary operation comparisons
 *
 * @param data          [ShadowedDataT] The data to sort
 * @param ascending     [bool]          Flag to indicate the sorting order
 * @param comp          [CompT]         The binary operator object
 */
template<typename ShadowedDataT, typename CompT>
void elbowSortCore(ShadowedDataT& data, bool ascending, CompT comp) noexcept {
    auto& active = data.getActive(); // Get the source vector (the data to sort)
    auto& shadow = data.getShadow(); // Get the target vector (the sorted data)

    size_t N = data.size();         // The total size is the same or both vectors
    size_t left = std::distance(
            active.begin(),
            (ascending) ?
                std::min_element(active.begin(), active.end()) :
                std::max_element(active.begin(), active.end())
    );                              // start 'left' from elbow of the bitonic
    size_t right = (left == N-1) ? 0 : left + 1;

    // Walk in opposite directions from elbow and insert-sort to target vector
    for (size_t i = 0 ; i<N ; ++i) {
        if (comp(active[left], active[right])) {
            shadow[i] = active[left];
            left = (left == 0) ? N-1 : left -1; // cycle decrease
        }
        else {
            shadow[i] = active[right];
            right = (right + 1) % N;            // cycle increase
        }
    }
    data.switch_active();           // Switch active-shadow buffers
}

/*!
 * Sort a shadowed buffer using the "elbow sort" algorithm.
 *
 * @tparam ShadowedDataT    A Shadowed buffer type with random access iterator.
 *
 * @param data          [ShadowedDataT] The data to sort
 * @param ascending     [bool]          Flag to indicate the sorting order
 */
template<typename ShadowedDataT>
void elbowSort(ShadowedDataT& data, bool ascending) noexcept {
    if (ascending)
        elbowSortCore(data, ascending, std::less<>());
    else
        elbowSortCore(data, ascending, std::greater<>());
}


/*!
 * Takes two sequences and selects either the larger or the smaller items
 * in one-to-one comparison between them. If the initial sequences are bitonic, then
 * the result is a bitonic sequence too!
 *
 * @tparam ValueT   The underlying type of the sequences
 *
 * @param local     [ValueT*]       Pointer to the local sequence
 * @param remote    [const ValueT*] Pointer to the remote sequence (copied locally by MPI)
 * @param count     [size_t]        The number of items to process
 * @param keepSmall [bool]          Flag to indicate if we keep the small items in local sequence
 */
template<typename ValueT>
void keepMinOrMax(ValueT* local, const ValueT* remote, size_t count, bool keepSmall) noexcept {
    std::transform(
            local, local + count,
            remote,
            local,
            [&keepSmall](const ValueT& a, const ValueT& b){
                return (keepSmall) ? std::min(a, b) : std::max(a, b);
            });
}

/*
 * ============================== Sort algorithms ==============================
 */


/*!
 * A distributed version of the Bitonic sort algorithm.
 *
 * @note
 *  Each MPI process should run an instance of this function.
 *
 * @tparam ShadowedDataT    A Shadowed buffer type with random access iterator.
 *
 * @param data          [ShadowedDataT] The local to MPI process data to sort
 * @param Processes     [mpi_id_t]      The total number of MPI processes
 * @param rank          [mpi_id_t]      The current process id
 */
template<typename ShadowedDataT>
void distBitonic(ShadowedDataT& data) {
    // Initially sort to create a half part of a bitonic sequence
    timeCall(Timer_fullSort, fullSort, data, ascending(rank, 0));

    // Run through sort network using elbow-sort ( O(LogN * LogN) iterations )
    auto p = static_cast<uint32_t>(std::log2(Processes));
    for (size_t depth = 1; depth <= p; ++depth) {
        for (size_t step = depth; step > 0;) {
            --step;
            // Find out exchange configuration
            auto part = partner(rank, step);
            auto ks = keepSmall(rank, part, depth);
            // Exchange with partner, keep nim-or-max
            exchange(data, part, ks, tag);

        }
        // sort - O(N)
        timeCall(Timer_elbowSort, elbowSort, data, ascending(rank, depth));
    }
}

#endif //DISTBITONIC_H_
